from forge import Forge
import logging
import waiting

import pytest


_logger = logging.getLogger(__name__)


@pytest.fixture
def forge(request):

    returned = Forge()

    @request.addfinalizer
    def cleanup():
        returned.verify()
        returned.restore_all_replacements()

    return returned

@pytest.fixture
def timeline(forge):

    class VirtualTimeline(object):

        class FirstTestException(Exception):
            pass


        class SecondTestException(Exception):
            pass


        def __init__(self):
            super(VirtualTimeline, self).__init__()
            self.virtual_time = 0
            self.sleeps_performed = []
            self.predicate_satisfied = False
            self.satisfy_at_time = None
            self.satisfy_after_time = None
            self.predicate_sleep = 0


        def sleep(self, delta):
            self.sleeps_performed.append(delta)
            self.virtual_time += delta
            assert 1000 > len(self.sleeps_performed), 'Infinite loop'

        def time(self):
            return self.virtual_time

        def predicate(self):
            if self.satisfy_at_time is not None and self.satisfy_at_time == self.virtual_time:
                self.predicate_satisfied = True
            if self.satisfy_after_time is not None and self.satisfy_after_time <= self.virtual_time:
                self.predicate_satisfied = True
            self.virtual_time += self.predicate_sleep
            _logger.debug('Predicate: time is now %s', self.virtual_time)
            return self.predicate_satisfied

        def raising_predicate(self):
            if self.virtual_time == 0:
                raise self.FirstTestException()
            return self.predicate()

        def raising_two_exceptions_predicate(self):
            if self.virtual_time == 1:
                raise self.SecondTestException()
            return self.raising_predicate()


    returned = VirtualTimeline()

    forge.replace_with(waiting.time_module, "sleep", returned.sleep)
    forge.replace_with(waiting.time_module, "time", returned.time)
    forge.replace_with(waiting.deadlines.time_module, "time", returned.time)

    return returned

@pytest.fixture
def predicates(forge):
    return [forge.create_wildcard_mock() for i in range(5)]



# coding=utf-8
import pytest
import waiting
from waiting.exceptions import IllegalArgumentError, TimeoutExpired


@pytest.mark.parametrize('arg', [
    ([1, 2, 4, 8, 16, 32, 64, 100, 100], (1, 100, 2)),
    ([1, 3, 9, 27, 81, 101, 101, 101], (1, 101, 3)),
    ([1, 2, 4, 8, 16, 32, 64, 128, 201], (1, 201)),
    ([1, 3, 9, 27, 81], (1, None, 3)),
    ([1, 2, 4, 8, 16, 32, 64, 128, 256, 512], (1, None)),
])
def test_exponential_sleep(timeline, arg):
    sleeps, sleep_seconds_arg = arg
    timeline.satisfy_after_time = sum(sleeps)
    waiting.wait(timeline.predicate, sleep_seconds=sleep_seconds_arg)
    assert sleeps == timeline.sleeps_performed

def test_waiting_does_not_expire(timeline):
    num_tries = 9
    sleep_seconds = 10
    timeline.satisfy_at_time = num_tries * sleep_seconds
    waiting.wait(timeline.predicate, sleep_seconds=sleep_seconds, timeout_seconds=(sleep_seconds*num_tries)+1)

def test_waiting_expires(timeline):
    with pytest.raises(waiting.TimeoutExpired):
        waiting.wait(timeline.predicate, sleep_seconds=3, timeout_seconds=5) # timeout_seconds intentionally not divided by sleep
    assert timeline.virtual_time == 5

def test_iterwait_sleep_time(timeline):
    sleep_seconds = 5
    num_sleeps = 7
    timeline.satisfy_at_time = sleep_seconds * num_sleeps
    previous_time = timeline.virtual_time
    for retry in waiting.iterwait(timeline.predicate, sleep_seconds=sleep_seconds):
        assert timeline.virtual_time == previous_time
        previous_time = timeline.virtual_time + sleep_seconds
        timeline.sleep(1.5) # shouldn't affect the sleep
    assert timeline.virtual_time == (num_sleeps * sleep_seconds)

def test__wait_any(predicates, forge, timeline):
    for p in predicates:
        p().and_return(False)
    predicates[0]().and_return(True)
    forge.replay()
    waiting.wait(waiting.ANY(predicates))

def test_wait_any_satisfied_on_start(predicates, forge, timeline):
    for index, p in enumerate(predicates):
        works = (index == 2)
        p().and_return(works)
        if works:
            break
    forge.replay()
    waiting.wait(waiting.ANY(predicates))
    assert timeline.virtual_time == 0

def test_wait_all(predicates, forge, timeline):
    for iteration in range(len(predicates) * 2, 0, -1):
        iteration -= 1
        for index, p in reversed(list(enumerate(predicates))):
            if index > iteration:
                continue
            p().and_return(index == iteration)
    forge.replay()
    waiting.wait(waiting.ALL(predicates))
    assert timeline.virtual_time == ((len(predicates) * 2) - 1)


def test_no_handled_exceptions(timeline):
    with pytest.raises(timeline.FirstTestException):
        waiting.wait(timeline.raising_predicate)

def test_another_handled_exceptions(timeline):
    with pytest.raises(timeline.SecondTestException):
        waiting.wait(timeline.raising_two_exceptions_predicate,
                     expected_exceptions=timeline.FirstTestException)

def test_one_class(timeline):
    timeline.satisfy_at_time = 3
    waiting.wait(timeline.raising_predicate, expected_exceptions=timeline.FirstTestException)
    assert timeline.virtual_time == 3

def test_correct_tuple(timeline):
    timeline.satisfy_at_time = 3
    waiting.wait(timeline.raising_two_exceptions_predicate,
                 expected_exceptions=(timeline.FirstTestException, timeline.SecondTestException))
    assert timeline.virtual_time == 3

def test_none(timeline):
    with pytest.raises(IllegalArgumentError):
        waiting.wait(timeline.raising_predicate, expected_exceptions=None)

def test_non_tuple(timeline):
    with pytest.raises(IllegalArgumentError):
        waiting.wait(timeline.raising_predicate, expected_exceptions='Not a tuple')


def test_unicode_logging():
    unicode_string = u'symbols 올 д'
    timeout = 3
    instance = TimeoutExpired(timeout, unicode_string)
    assert u'{0}'.format(instance) == u'Timeout of {0} seconds expired waiting for {1}'.format(timeout, unicode_string)


@pytest.mark.parametrize(['exception', 'expected_exceptions'], [
                         (StopIteration, ()),
                         (TimeoutExpired, StopIteration)])
def test_stop_iteration_inside_predicate(exception, expected_exceptions):
    def predicate():
        raise StopIteration()
    with pytest.raises(exception):
        waiting.wait(predicate, timeout_seconds=0,
                     expected_exceptions=expected_exceptions)


def test_nexted_stop_iteration_preserve_exception():
    expected_exception = StopIteration('foo')

    def predicate():
        raise expected_exception

    with pytest.raises(StopIteration) as ex:
        waiting.wait(predicate, timeout_seconds=0)
    assert ex.value is expected_exception


def test_nexted_stop_iteration_preserve_traceback():
    expected_exception = StopIteration('foo')

    def predicate():
        raise expected_exception

    with pytest.raises(StopIteration) as ex:
        waiting.wait(predicate, timeout_seconds=0)
    assert '        raise expected_exception' in ex.traceback[-1].statement.lines



import waiting

import pytest


@pytest.mark.parametrize('do_sleep', [True, False])
def test_predicate_sleeps_coelesce(timeline, do_sleep):
    timeline.predicate_sleep = 0.5 if do_sleep else 0
    sleep_seconds = 2
    expected_sleep = sleep_seconds - timeline.predicate_sleep
    timeline.satisfy_after_time = 10
    waiting.wait(timeline.predicate, sleep_seconds=2)
    assert timeline.sleeps_performed == [expected_sleep for i in range(5)]



try:
    from flux import current_timeline as time_module
except ImportError:
    import time as time_module


class Deadline(object):

    def is_expired(self):
        raise NotImplementedError()  # pragma: no cover

    def get_num_seconds_remaining(self):
        return None


class Within(Deadline):

    def __init__(self, seconds):
        super(Within, self).__init__()
        self._deadline = time_module.time() + seconds

    def is_expired(self):
        return time_module.time() >= self._deadline

    def get_num_seconds_remaining(self):
        return self._deadline - time_module.time()


class Whenever(Deadline):

    def is_expired(self):
        return False


def make_deadline(seconds):
    if seconds is None:
        return Whenever()
    return Within(seconds)



import inspect
import sys

from .__version__ import __version__
from contextlib import contextmanager

try:
    from flux import current_timeline as time_module
except ImportError:
    import time as time_module
from .deadlines import make_deadline as _make_deadline
from .exceptions import TimeoutExpired, IllegalArgumentError, NestedStopIteration


def wait(*args, **kwargs):
    result = _Result()
    try:
        for x in iterwait(result=result, *args, **kwargs):
            pass
    except NestedStopIteration as e:
        e.reraise()
    return result.result


def iterwait(predicate, timeout_seconds=None, sleep_seconds=1, result=None, waiting_for=None,
             on_poll=None, expected_exceptions=()):

    if not isinstance(expected_exceptions, tuple) and not (isinstance(expected_exceptions, type)
                                                           and issubclass(expected_exceptions, Exception)):
        raise IllegalArgumentError(
            'expected_exceptions should be tuple or Exception subclass')
    if on_poll is not None and not callable(on_poll):
        raise IllegalArgumentError(
            'on_poll should be callable')
    timeout = _make_deadline(timeout_seconds)
    if result is None:
        result = _Result()
    if waiting_for is None:
        waiting_for = str(predicate)
    sleep_generator = _get_sleep_generator(timeout, sleep_seconds)
    while True:
        with _end_sleeping(next(sleep_generator)) as cancel_sleep:
            try:
                result.result = predicate()
                if on_poll is not None:
                    on_poll()
            except expected_exceptions:
                pass
            except StopIteration:
                exc_info = sys.exc_info()
                raise NestedStopIteration(exc_info)
            if result.result:
                cancel_sleep()
                return
            if timeout.is_expired():
                raise TimeoutExpired(timeout_seconds, waiting_for)
            yield


@contextmanager
def _end_sleeping(total_seconds):
    deadline = _make_deadline(total_seconds)
    sleep_toggle = _SleepToggle()
    yield sleep_toggle
    if sleep_toggle.enabled:
        time_module.sleep(max(0, deadline.get_num_seconds_remaining()))

class _SleepToggle(object):
    enabled = True

    def __call__(self):
        self.enabled = False


class _Result(object):
    result = None


def _get_sleep_generator(timeout, sleep_seconds):
    if type(sleep_seconds) not in (tuple, list):
        sleep_seconds = (sleep_seconds, sleep_seconds, 1)
    if len(sleep_seconds) <= 2:
        sleep_seconds = (sleep_seconds[0], sleep_seconds[1], 2)
    elif len(sleep_seconds) < 2:
        sleep_seconds = (sleep_seconds[0], None, 2)
    current_sleep, end_sleep, multiplier = sleep_seconds
    while True:
        time_remaining = timeout.get_num_seconds_remaining()
        if time_remaining is not None:
            current_sleep = min(time_remaining, current_sleep)
        current_sleep = max(0, current_sleep)
        yield current_sleep
        current_sleep *= multiplier
        if end_sleep is not None:
            current_sleep = min(end_sleep, current_sleep)


class Aggregate(object):

    def __init__(self, predicates):
        super(Aggregate, self).__init__()
        self.predicates = list(predicates)


class ANY(Aggregate):

    def __call__(self):
        return any(p() for p in self.predicates)


class ALL(Aggregate):

    def __call__(self):
        for index in range(len(self.predicates), 0, -1):
            index -= 1
            if self.predicates[index]():
                self.predicates.pop(index)
        return not self.predicates



import os
from setuptools import setup, find_packages

with open(os.path.join(os.path.dirname(__file__), "waiting", "__version__.py")) as version_file:
    exec(version_file.read())

setup(name="waiting",
      classifiers = [
          "Programming Language :: Python :: 2.6",
          "Programming Language :: Python :: 2.7",
          "Programming Language :: Python :: 3.3",
          "Programming Language :: Python :: 3.4",
          "Programming Language :: Python :: 3.5",
          ],
      description="Utility for waiting for stuff to happen",
      license="BSD",
      author="Rotem Yaari",
      author_email="vmalloc@gmail.com",
      version=__version__,
      packages=find_packages(exclude=["tests"]),
      install_requires=[],
      tests_require=["pyforge"],
      scripts=[],
      namespace_packages=[]
      )



import sys

class TimeoutExpired(Exception):
    def __init__(self, timeout_seconds, what):
        super(TimeoutExpired, self).__init__(timeout_seconds, what)
        self._timeout_seconds = timeout_seconds
        self._what = what
    def __str__(self):
        return "Timeout of {0} seconds expired waiting for {1}".format(self._timeout_seconds, self._what)
    def __repr__(self):
        return "{0}: {1}".format(type(self).__name__, self)
    def __unicode__(self):
        return u"Timeout of {0} seconds expired waiting for {1}".format(self._timeout_seconds, self._what)


class IllegalArgumentError(ValueError):
    pass


class NestedStopIteration(Exception):

    def __init__(self, exc_info):
        self.exc_info = exc_info

    def reraise(self):
        if sys.version_info[0] == 3:
            self._reraise3()
        else:
            self._reraise2()

    def _reraise2(self):
        exec("raise self.exc_info[0], self.exc_info[1], self.exc_info[2]")

    def _reraise3(self):
        _, exc_value, tb = self.exc_info
        raise exc_value.with_traceback(tb)
